/*
 * $Id: person.c,v 0.31 1995/01/01 19:31:47 ceder Exp $
 * Copyright (C) 1991, 1992, 1993, 1994, 1995  Lysator Academic Computer Association.
 *
 * This file is part of the LysKOM server.
 * 
 * LysKOM is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 1, or (at your option) 
 * any later version.
 * 
 * LysKOM is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with LysKOM; see the file COPYING.  If not, write to
 * Lysator, c/o ISY, Linkoping University, S-581 83 Linkoping, SWEDEN,
 * or the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, 
 * MA 02139, USA.
 *
 * Please mail bug reports to bug-lyskom@lysator.liu.se. 
 */
/*
 * person.c
 *
 * All atomic calls that deals with persons.
 */

static char *rcsid = "$Id: person.c,v 0.31 1995/01/01 19:31:47 ceder Exp $";
#include "rcs.h"
USE(rcsid);

#include <limits.h>
#include <stdio.h>
#ifdef HAVE_STDLIB_H
#  include <stdlib.h>
#endif
#ifdef HAVE_STRING_H
#  include <string.h>
#endif
#include <time.h>
#include <unistd.h>		/* for crypt() on Linux */
#include <setjmp.h>
#include <sys/types.h>
#ifdef HAVE_STDARG_H
#    include <stdarg.h>
#endif

#include "misc-types.h"
#include "s-string.h"
#include "kom-types.h"
#include "services.h"
#include "manipulate.h"
#include "string-malloc.h"
#include "debug.h"
#include "cache.h"
#include "kom-errno.h"
#include "com.h"
#include "connections.h"
#include "server/smalloc.h"
#include "config.h"
#include "log.h"
#include "minmax.h"
#include "lyskomd.h"
#include "param.h"

/*
 * Static functions.
 */

static Bool
legal_passwd(const String pwd)
{
    int i;

    for (i = 0; i < s_strlen(pwd); i++)
	if (pwd.string[i] == '\0')
	    return FALSE;

    return TRUE; 
}

static Success
do_set_passwd( Password        pwd,
	       const String   new_pwd)
{
#ifdef ENCRYPT_PASSWORDS
    char salt[3];
    static char crypt_seed[] = 
      "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789./";
    char *password;

    if ( !legal_passwd(new_pwd) )
	return FAILURE;

    salt[0] = crypt_seed [rand() % (sizeof (crypt_seed) - 1)];
    salt[1] = crypt_seed [rand() % (sizeof (crypt_seed) - 1)];
    salt[2] = '\0';

    password = s_crea_c_str(new_pwd);
    if (password == NULL)
	return FAILURE;
    else
    {
	strcpy((char *)pwd, (const char *)crypt(password, salt));
	string_free(password);
	return OK;
    }

#else

    /* Compatibility mode... */
    if ( !legal_passwd(new_pwd) )
	return FAILURE;

    *pwd++ = new_pwd.len;		/* Indeed too easy crypt... */
    strncpy(pwd, (const char *)new_pwd.string,
	    min( PASSWD_LEN-1, new_pwd.len ));

    return OK;

#endif
}



/*
 * Mark a text. No check is done if pers_p is really allowed to mark it.
 * This function can mark a text with mark_type == 0.  If a previous mark
 * existed it is changed to the given mark_type.
 */

static Success
do_mark_text(Pers_no    pers_no,
	     Person    *pers_p,		/* May be NULL */
	     Text_no    text_no,
	     Text_stat *text_s,		/* May be NULL */
	     u_char     mark_type)
{
    Mark      *markp;
    Mark      *old_mark = NULL;
    int        i;
    BUGDECL;

    if ( pers_p == NULL )
	GET_P_STAT(pers_p, pers_no, FAILURE);
	
    /* Locate the mark. */
    
    for ( i = pers_p->marks.no_of_marks, markp = pers_p->marks.marks;
	 i > 0 && old_mark == NULL;
	 i--, markp++ )
    {
	if ( markp->text_no == text_no )
	    old_mark = markp;
    }

    if (old_mark != NULL)
    {
	/* Change marktype of an already marked text. (No need to check
	   that the text exists). */
	BUG(("do_mark_text(): Change type of mark.\n"));
	
	old_mark->mark_type = mark_type;
	mark_person_as_changed(pers_no);
    }
    else
    {
	/* A new mark. Check that the text exists. */
	BUG(("do_mark_text(): new mark.\n"));

	if ( text_s == NULL )
	    GET_T_STAT(text_s, text_no, FAILURE);

	if (text_s->no_of_marks >= param.max_marks_text
	    || pers_p->marks.no_of_marks >= param.max_marks_person)
	{
	    kom_errno = KOM_MARK_LIMIT;
	    return FAILURE;
	}

	text_s->no_of_marks++;
	mark_text_as_changed(text_no);
	
	pers_p->marks.marks = srealloc(pers_p->marks.marks,
			       ++pers_p->marks.no_of_marks * sizeof(Mark));
	pers_p->marks.marks[pers_p->marks.no_of_marks - 1].text_no = text_no;
	pers_p->marks.marks[pers_p->marks.no_of_marks - 1].mark_type
	    = mark_type;
	mark_person_as_changed(pers_no);
    }
    
    return OK;
}

/* Unmark a text. Returns a failure if the text was not marked. */

static Success
do_unmark_text(Pers_no    pers_no,
	       Person    *pers_p, /* May be NULL */
	       Text_no    text_no,
	       Text_stat *text_s) /* May be NULL */
{
    Mark      *markp;
    Mark      *old_mark = NULL;
    int        i;
    BUGDECL;

    if ( pers_p == NULL )
	GET_P_STAT(pers_p, pers_no, FAILURE);
	
    /* Locate the mark. */
    
    for ( i = pers_p->marks.no_of_marks, markp = pers_p->marks.marks;
	 i > 0 && old_mark == NULL;
	 i--, markp++ )
    {
	if ( markp->text_no == text_no )
	    old_mark = markp;
    }

    if (old_mark != NULL)
    {
	/* Delete this mark. */
	BUG(("do_unmark_text(): Delete mark.\n"));
	
	while ( old_mark < ( pers_p->marks.marks
			    + pers_p->marks.no_of_marks - 1 ))
	{
	    *old_mark = *(old_mark + 1);
	    old_mark++;
	}

	pers_p->marks.no_of_marks--;
	mark_person_as_changed(pers_no);
	
	/* ...and don't forget to decrease the ref_cnt in the text_stat... */

	if ( (text_s = cached_get_text_stat(text_no)) != NULL )
	{
	    if ( text_s->no_of_marks == 0 )
	    {
		log("WNG: do_unmark_text(): Text %d has no_of_marks==0,\
 		     but person %d had marked the text.",
		    text_no, pers_no);
	    }
	    else
	    {
		text_s->no_of_marks--;
		mark_text_as_changed(text_no);
	    }
	}
	/* ...but it is not an error if someone has deleted the text. */
    }
    else
    {
	/* An attempt to delete a non-existent mark. */
	BUG(("do_unmark_text(): delete non-existent mark.\n"));
	kom_errno = KOM_NOT_MARKED;
	return FAILURE;
    }
    
    return OK;
}


/*
 * Change user_area of a person. If text_no is 0, there will be
 * no user_area.
 */

static Success
do_set_user_area(Pers_no     pers_no,
		 Person	   * pers_p, /* May be NULL */
		 Text_no     user_area)
{
    Text_stat * old_user_area;
    Text_stat * new_user_area = NULL; /* To stop gcc complaining. */

    /* Check that the new user_area exists before deleting the old*/

    if ( user_area != 0 )
    {
	GET_T_STAT(new_user_area, user_area, FAILURE);
	if ( new_user_area->no_of_marks >= param.max_marks_text )
	{
	    log("%s(%d, %d): New user_area's mark count (%d) > %d.\n",
		"LIMIT: set_user_area",
		pers_no, user_area, new_user_area->no_of_marks,
		param.max_marks_text);
	    kom_errno = KOM_MARK_LIMIT;
	    return FAILURE;
	}
    }

    if ( pers_p == NULL )
	GET_P_STAT(pers_p, pers_no, FAILURE);
    
    /* Unmark the previous user_area if it exists. */

    if ( pers_p->user_area != 0
	&& (old_user_area = cached_get_text_stat(pers_p->user_area)) != NULL)
    {
	if ( old_user_area->no_of_marks > 0 )
	{
	    --old_user_area->no_of_marks;
	    mark_text_as_changed( pers_p->user_area );
	}
	else
	{
	    log("ERROR: set_user_area(%d, %d): Old user_area %d not marked\n",
		pers_no, user_area, pers_p->user_area);
	}
    }

    /* Mark the new user_area */

    if ( user_area != 0 )
    {
	++new_user_area->no_of_marks;
	mark_text_as_changed( user_area );
    }
    
    pers_p->user_area = user_area;
    mark_person_as_changed (pers_no);

    return OK;
}
/*
 * End of static functions.
 */

/*
 * Functions that are exported to the server.
 */

/*
 * Delete a person. (The mailbox is not deleted).
 */

Success
do_delete_pers (Pers_no pers_no)
{
    Person * pers_p;
    int	     i;
    
    GET_P_STAT(pers_p, pers_no, FAILURE);

    /* Cancel membership in all confs. */
    /* Note that because of the way do_sub_member is written the */
    /* loop must be executed this way. */
    for ( i = pers_p->conferences.no_of_confs - 1; i >= 0; i-- )
    {
	if ( do_sub_member( pers_p->conferences.confs[ i ].conf_no, NULL, NULL,
			   pers_no, pers_p,
			   pers_p->conferences.confs + i) != OK)
	{
	    log("ERROR: do_delete_pers(): can't sub_member\n");
	}
    }

    for ( i = pers_p->marks.no_of_marks; i > 0; i-- )
    {
	if ( do_unmark_text(pers_no, pers_p,
			    pers_p->marks.marks[0].text_no, NULL) != OK )
	{
	    log("WNG: do_delete_pers(): can't unmark text %d (i == %d)\n",
		pers_p->marks.marks[0].text_no, i);
	}
    }
    
    if ( do_set_user_area(pers_no, pers_p, 0) != OK )
    {
	log("WNG: do_delete_pers(): can't unmark user_area\n");
    }

    s_clear( &pers_p->username );
    sfree( pers_p->created_texts.texts );
    pers_p->created_texts.texts = NULL;
    pers_p->created_texts.no_of_texts = 0;

    cached_delete_person(pers_no);
    /* ??? Eller r{cker det med cached_delete_conf() ??? */
    
    /* ASYNC */
    return OK;    
}

/*
 * Check a password
 */
Success
chk_passwd (Password      pwd,
	    const String  try)
{
#ifdef ENCRYPT_PASSWORDS
    char *c_try;

    c_try = s_crea_c_str(try);
    if (c_try == NULL)
	return FAILURE;
    
    if (strcmp ((const char *)pwd,
		(const char *)crypt(c_try, (const char *)pwd)) != 0)
    {
	string_free(c_try);
	return FAILURE;
    }
    else
    {
	string_free(c_try);
	return OK;
    }

#else

    int i;
    
    if ( try.len != *pwd )
	return FAILURE;

    for ( i = 0; i < try.len; i++ )
	if ( pwd[ i + 1 ] != try.string[ i ] )
	    return FAILURE;

    return OK;

#endif
}


/*
 * Mark a text.
 *
 * If the you already have marked the text the mark-type will be changed. 
 * If mark_type == 0 the mark will be deleted. You can only mark texts
 * that you are allowed to read. You are always allowed to read all
 * texts that you have marked even if you are no longer a recipient
 * of the text.
 */

extern Success
mark_text_old(Text_no text_no,
	      u_char mark_type)
{
    if (mark_type == 0)
	return unmark_text(text_no);
    else
	return mark_text(text_no, mark_type);
}

extern Success
mark_text(Text_no text_no,   /* Will fail if the user is not */
	  u_char mark_type)	/* allowed to read the text.	*/
{
    Text_stat *text_s = NULL;
    BUGDECL;
        
    CHK_LOGIN(FAILURE);

    /* Check that the user is allowed to read the text. */
    GET_T_STAT(text_s, text_no, FAILURE);
    if ( !text_read_access(text_no, text_s) )
    {
	kom_errno = KOM_PERM;
	return FAILURE;
    }

    BUG(("Person %d markerar text %lu med typ %d\n",
	 ACTPERS, text_no, mark_type));
    
    return do_mark_text(ACTPERS, ACT_P, text_no, text_s, mark_type);
}

extern Success
unmark_text(Text_no text_no)
{
    Text_stat *text_s = NULL;
    BUGDECL;
        
    CHK_LOGIN(FAILURE);
    BUG(("Person %d avmarkerar text %lu\n", ACTPERS, text_no));
    
    return do_unmark_text(ACTPERS, ACT_P, text_no, text_s);
}

/*
 * Create a new person.
 *
 * If parameter ``Anyone can create new persons'' is true
 * anyone can create themself. Otherwise you must be logged in to
 * create a new person, and you must have the ``create_pers'' bit
 * set, or ``Anyone can create_new_persons'' must be set.
 *
 * If you are not logged in an auto-login as the created person is
 * performed. (This is so that it is impossible to create a million
 * persons (and so destroy the database) anonymously. This way it will
 * at least be possible to see from which machine the attack came).
 *
 * If you are logged in you will be supervisor and creator of the new
 * person, otherwise the new person will be supervisor and creator.
 *
 * The new person will be rd_prot, but not secret. He will have no
 * privileges, except maybe change_name (if ``Default change name
 * capability'' is true).
 *
 * This function returns the Pers_no of the new person, or 0 if an error
 * occured. (0 is an illegal Pers_no).
 */
extern Pers_no
create_person (const String  name,
	       const String  passwd )
{
    Pers_no 	  new_user;
    Conference	* mailbox;
    Person	* pers_p;
    

    if (ACTPERS == 0)
    {
	if (!param.create_person_before_login && ACTPERS == 0 )
	{
	    kom_errno = KOM_LOGIN;
	    return 0;
	}
    }
    else
    {
	if (param.anyone_can_create_new_persons == FALSE
	    && !ENA(create_pers, 0) )
	{
	    kom_errno = KOM_PERM;
	    return 0;
	}
    }

#if 0
    if ( !logins_allowed  && !ACTPERS ) /* A new user can't create himself */
    {					/* if he isn't allowed to login. */
	kom_errno = KOM_LOGIN_DISALLOWED;
	return 0;
    }
#endif

    if ( !legal_name( name ) )
    {
	/* kom_errno will be set by legal_name(). */
	return 0;
    }

    if ( !unique_name( name, 0 ) )
    {
	kom_errno = KOM_CONF_EXISTS;
	return 0;
    }

    if ( !legal_passwd( passwd ) )
    {
	kom_errno = KOM_PWD;
	return 0;
    }

    new_user = cached_create_conf( name );

    if ( (mailbox = cached_get_conf_stat( new_user ) ) == NULL)
    {
	restart_kom("create_person() - can't get conf_stat");
    }
    
    
    mailbox->creator		= ACTPERS ? ACTPERS : new_user;
    mailbox->creation_time 	= time(NULL);
    mailbox->presentation	= 0;		/* No presentation yet */
    mailbox->supervisor		= ACTPERS ? ACTPERS : new_user;
    mailbox->permitted_submitters = 0;
    mailbox->super_conf		= ACTPERS;
    mailbox->type.rd_prot	= 1;
    mailbox->type.original	= 0;
    mailbox->type.secret	= 0;
    mailbox->type.letter_box	= 1;
    mailbox->last_written	= mailbox->creation_time;
    mailbox->msg_of_day		= 0;
    mailbox->nice		= param.default_nice;

    mark_conference_as_changed (new_user);

    /* allocate Person */

    if ( cached_create_person( new_user ) == FAILURE )
    {
	cached_delete_conf( new_user );
	mark_conference_as_changed( new_user ); /*&*/
	return 0;
    }

    if ( ( pers_p = cached_get_person_stat( new_user )) == NULL )
    {
	cached_delete_conf( new_user );
	mark_conference_as_changed( new_user ); /*&*/
	restart_kom("create_person() - can't get pers_stat");
    }

	
    /* Fill in Person */
    if ( do_set_passwd( pers_p->pwd, passwd) != OK )
	restart_kom("create_person(): can't set passwd\n");

    pers_p->privileges.change_name = param.default_change_name;

    mark_person_as_changed( new_user );

    do_add_member( new_user, mailbox, new_user, pers_p, UCHAR_MAX, 0 );

    if ( ACTPERS )
    {
	ACT_P->created_persons++;
	mark_person_as_changed( ACTPERS );
    }
    else
    {
	/* Auto login, always visible */
	login (new_user, passwd, FALSE);
    }

    return new_user;
}


/*
 * Get the person status of PERSON.
 */
extern  Success
get_person_stat (Pers_no	  person,
 		 Person		* result)
{
    Person *p_orig;
    Conference *pers_c;
    Access acc;
        
    GET_P_STAT(p_orig, person, FAILURE);
    GET_C_STAT(pers_c, person, FAILURE);
    
    acc = access_perm (person, pers_c, ACTPERS, ACT_P);
  
    if ( acc == error )
	return FAILURE;

    if ( acc == none )
    {
	kom_errno = KOM_UNDEF_PERS;
	return FAILURE;
    }

    *result = *p_orig;	/* Make a copy of the struct. */
    
    /* The user area is normally secret. */

    if( acc != unlimited )
	result->user_area	= 0;
    
    return OK;
}

/* As get_person_stat, but only return the name if bit 0 in mask is set. */

extern  Success
get_person_stat_old (Pers_no	  person,
		     int 	  mask,
		     Person	* result)
{
    Person *p_orig;
    Conference *pers_c;
    Access acc;
        
    CHK_LOGIN(FAILURE);
    
    GET_P_STAT(p_orig, person, FAILURE);
    GET_C_STAT(pers_c, person, FAILURE);
    
    acc = access_perm (person, pers_c, ACTPERS, ACT_P);
  
    if ( acc == error )
	return FAILURE;

    if ( acc == none )
    {
	kom_errno = KOM_UNDEF_PERS;
	return FAILURE;
    }

    *result = *p_orig;	/* Make a copy of the struct. */
    
    /* The user area is normally secret. */

    if( acc != unlimited )
	result->user_area	= 0;

    if ( !(mask & 1 ) )
	result->username = EMPTY_STRING;
    
    return OK;
}


/* /// */
extern  Success
get_created_texts(Pers_no       pers_no,
		  Local_text_no first,
		  u_long        len,
		  Text_list   * result)
{
    Person *pers_p;
    Access acc;
    Local_text_no new_first;

    CHK_LOGIN(FAILURE);

    GET_P_STAT(pers_p, pers_no, FAILURE);

    acc = access_perm(pers_no, NULL, ACTPERS, ACT_P);

    if ( acc == error )
	return FAILURE;

    if ( acc == none )
    {
	kom_errno = KOM_UNDEF_PERS;
	return FAILURE;
    }

    *result = pers_p->created_texts;
    new_first = max(result->first_local_no, first);

    if ( new_first >= result->first_local_no + result->no_of_texts)
    {
	kom_errno = KOM_NO_SUCH_LOCAL_TEXT;
	return FAILURE;
    }

    result->texts += new_first - result->first_local_no;
    result->no_of_texts = min(len, (result->no_of_texts
				    - (new_first - result->first_local_no)));
    result->first_local_no = new_first;

    return OK;
}



/*
 * Set privilege bits of a person. You must have the wheel bit set
 * to be allowed to do this.
 */
extern Success
set_priv_bits(	Pers_no		person,
		Priv_bits	privileges )
{
    Person *p;
    
    CHK_LOGIN(FAILURE);

    GET_P_STAT(p, person, FAILURE);

    if ( ENA(wheel, 6) )
    {
	p->privileges = privileges;
	mark_person_as_changed( person );
	return OK;
    }
    else
    {
	kom_errno = KOM_PERM;
	return FAILURE;
    }
}

/*
 * Set password. You may set the password of yourself and all persons that
 * you are supervisor of. OLD_PWD is your password. 
 */
extern Success 
set_passwd (Pers_no	    person,
	    const String   old_pwd,	/* of the one who is changing the pwd,
					   not necessarily the person whose pwd
					   is changed. */
	    const String   new_pwd)	/* of person */
{
    Person *p;
    
    CHK_LOGIN(FAILURE);

    GET_P_STAT(p, person, FAILURE);

    if ( person != ACTPERS && !ENA(wheel, 7)
	&& !is_supervisor (person, NULL, ACTPERS, ACT_P) )
    {
	kom_errno = KOM_PERM;
	return FAILURE;	   /* Not allowed to change the other persons pwd */
    }

    if ( chk_passwd( ACT_P->pwd, old_pwd ) == FAILURE )     
    {
	kom_errno = KOM_PWD;
	return FAILURE;
    }

    if ( do_set_passwd( p->pwd, new_pwd ) != OK)
    {
	kom_errno = KOM_PWD;
	return FAILURE;
    }

    mark_person_as_changed( person );

    return OK;
}

/*
 * Ask which texts a person has read in a certain conference.
 * Can be done before loggin. Will return EMPTY_MEMBERSHIP if VICTIM
 * has his unread_is_secret-flag set. This can not be distinguished
 * from the case when VICTIM has not read any texts in CONF_NO. This
 * is a feature, not a bug. (At least I think so. /ceder+++)
 *
 * (*** Well, I've changed my mind. This call is full of security holes.
 *	/ceder)
 *
 * Will fail if VICTIM is not a member of CONF_NO.
 */
extern Success
query_read_texts(Pers_no	  victim,
		 Conf_no	  conf_no,
		 Membership * result)	/* Points to area to store result in */
{
    Person	* victim_p;
    Membership  * membp;
    Access	  victim_acc;

    GET_P_STAT( victim_p, victim, FAILURE);

    victim_acc = access_perm(victim, NULL, ACTPERS, ACT_P);
    if ( victim_acc <= none )
    {
	kom_errno = KOM_UNDEF_PERS;
	return FAILURE;
    }

    if ( access_perm(conf_no, NULL, ACTPERS, ACT_P) <= none )
    {
	kom_errno = KOM_UNDEF_CONF;
	return FAILURE;
    }
    
    if ( (membp = locate_membership (conf_no, victim_p)) == NULL)
    {
	kom_errno = KOM_NOT_MEMBER;
	return FAILURE;
    }

    *result = *membp;

    if ( victim_p->flags.unread_is_secret &&
	victim_acc != unlimited)
    {
	result->last_time_read = NO_TIME;
	result->last_text_read = 0;
	result->no_of_read = 0;
	result->read_texts = NULL;
    }

    return OK;
}


/*
 * Set user_area. The text is a text which has previously been created
 * with create_text(). It typically contains option-settings for the
 * clients, e. g. if you want to be interrupted when new mail arrives.
 * The format of this text is not yet defined.
 *
 * set_user_area(0) to clear the user area.
 */

extern Success
set_user_area(Pers_no	pers_no,
	      Text_no	user_area)
{
    Person *pers_p;

    CHK_LOGIN(FAILURE);
    GET_P_STAT(pers_p, pers_no, FAILURE);

    if ( access_perm(pers_no, NULL, ACTPERS, ACT_P) != unlimited )
    {
	kom_errno = KOM_PERM;
	return FAILURE;
    }

    return do_set_user_area (pers_no, pers_p, user_area);
}
